/*
 * @(#)file      EngineIdGenerator.java
 * @(#)author    Sun Microsystems, Inc.
 * @(#)version   1.10
 * @(#)lastedit  04/04/21
 *
 * Copyright 2004 Sun Microsystems, Inc. All rights reserved.
 * SUN PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

//A single Java DMK import to access engine id toolkit
import com.sun.management.snmp.SnmpEngineId;

//java imports
import java.io.InputStreamReader;
import java.io.BufferedReader;

/**
 * This class lets you generate new SNMP engine ids.
 * This example makes an intensive use of the JDMK class 
 * {@link com.sun.management.snmp.SnmpEngineId}. 
 *
 * <p>An engine id uniquely identifies an SNMP entity. In the context of 
 * Java DMK an engine id is usually associated to a 
 * {@link com.sun.management.comm.SnmpV3AdaptorServer} or to a 
 * {@link com.sun.management.snmp.manager.SnmpSession}.
 *
 * <p>The engine id computation is compliant with SNMP V3 RFC 2571.
 * You can create an SnmpEngineId object in multiple ways:
 * <br>     * Providing an already computed engineId.
 * <br>     * Providing IP host name and (or) udp port.
 * <br>     * Providing IANA number.
 * <br>     * Current time based.
 *
 * <p>A useful SnmpEngineId feature is the ability to generate an SNMP OID
 * based on the engine id it models. It is useful when computing USM 
 * MIB user table indexes.
 * <p> Launch the example : java EngineIdGenerator
 * <br> Follow the instructions. Type h for help.
 */
public class EngineIdGenerator {
    public static void main(String[] args) {
        //Infinite loop to parse user input
        InputStreamReader i = new InputStreamReader(java.lang.System.in);
        BufferedReader in = new BufferedReader(i);
        SnmpEngineId engineId = null;
        print("Start making your engine id construct: (h for help)\n");
        while (true) {
            print("#:");
            try{
                String req = in.readLine();
                if (req.equals("oid")) {
                    if (engineId != null)
                        println("Generated OID       ******** [" + 
                                engineId.toOid() + "] ********** " );
                    else
                        println("WARNING : No engine id generated." +
                                " Generate one first.\n");
                    continue;
                }
                if (req.equals("exit")) java.lang.System.exit(0);
                if (req.equals("h") || req.equals("help")) { 
                    printHelp();
                    continue;
                }

                // Simply call static SnmpEngineId.createEngineId 
                // SnmpEngine Id factory method.
                //
                if (req.equals(""))
                    engineId = SnmpEngineId.createEngineId();
                else
                    engineId = SnmpEngineId.createEngineId(req.trim());
                
                println("Generated engine id ******** [" + 
                        engineId.toString() + "] ******** " );
            } catch (Exception e) {
                println("Badly formed input!!!");
                continue;
            }
        }
    }

    private static void printHelp() {
        println("**********************************" + " HELP " +
                "***********************************\n");
        println("Type a string in order to compute an SnmpEngineId. " +
                "Address means IP address or DNS name. Allowed formats:\n");
        println("\t * <address>:<port>:<IANA number>  " +
                "All these parameters are used to generate the id. " +
                "eg : localhost:8085:10\n");
        println("\t * <address>:<port> The IANA number will be the " +
                "Sun Microsystems one (42). eg : localhost:8085\n");
        println("\t * address  The port 161 will be used to generate " +
                "the engine id. The IANA number will be the Sun Microsystems " +
                "one (42). eg : localhost\n");
        println("\t * :port  The host to use is localhost. " +
                "The IANA number will be the Sun Microsystems one (42). " +
                "eg : :8085\n");
        println("\t * ::<IANA number>  The port 161 and localhost " +
                "will be used to generate the engine id. eg : ::10\n");
        println("\t * :<port>:<IANA number>  The host to use is localhost."+
                " eg : :8085:10\n");
        println("\t * <address>::<IANA number>  The port 161 will be used" +
                " to generate the engine id. eg : 127.0.0.1::10 \n");
        println("\t * :: The port 161, localhost and the Sun Microsystems' " +
                "IANA number will be used to generate the engine id. " +
		"eg : ::\n");
        println("\n--- or ---\n");
        println("Type 'oid' in order to generate the OID representation " +
                "of the previously calculated engine id. " +
                "It is useful when dealing with USM MIB.");
        println("\n--- or ---\n");
        println("Press <RETURN> in order to generate a current time based " +
		"engine id.");
        println("\n--- or ---\n");
        println("Type 'exit' to exit.");

        println("\n**********************************" + " HELP " +
                "**********************************\n");
    }

    /**
     * print/println stuff...
     */
    private final static void println(String msg) {
        java.lang.System.out.println(msg);
    }

    private final static void print(String msg) {
        java.lang.System.out.print(msg);
    }
}
