/*
 * @(#)file      JmxMBeanEntryImpl.java
 * @(#)author    Sun Microsystems, Inc.
 * @(#)version   1.8
 * @(#)lastedit  04/04/21
 * 
 * Copyright 2004 Sun Microsystems, Inc. All rights reserved.
 * SUN PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

// java imports
//
import java.io.Serializable;
import java.util.Enumeration;
import java.util.List;
import java.util.ArrayList;

// jmx/jdmk imports
//
import com.sun.management.snmp.SnmpStatusException;
import javax.management.MBeanServer;
import com.sun.management.snmp.SnmpInt;
import com.sun.management.snmp.SnmpString;
import com.sun.management.snmp.SnmpOid;
import com.sun.management.snmp.SnmpVarBind;
import javax.management.ObjectName;
import javax.management.MBeanServerNotification;
import javax.management.MBeanServerConnection;
import javax.management.Notification;
import javax.management.MBeanNotificationInfo;
import javax.management.MBeanInfo;
import javax.management.MBeanAttributeInfo;
import javax.management.NotificationEmitter;
import javax.management.NotificationListener;
import javax.management.NotificationFilter;
import javax.management.NotificationBroadcasterSupport;
import javax.management.InstanceAlreadyExistsException;
import javax.management.InstanceNotFoundException;
import javax.management.MalformedObjectNameException;
import com.sun.management.snmp.agent.SnmpMibSubRequest;
import com.sun.management.snmp.EnumRowStatus;
import com.sun.management.snmp.agent.SnmpMib;

/**
 * The class is used for instrumenting the "JmxMBeanEntry" entries.
 * It extends the mibgen generated JmxMBeanEntry skeleton.
 * An instance of this class mirrors an MBean registered in the
 * MBeanServer served by this MIB.
 *
 * This class instrument entries of the jmxMBeanTable table defined in the 
 * MIB. A JmxMBeanEntry represents an MBean registered in the exposed 
 * MBeanServer.
 * JmxMBeanEntry objects are created on the fly when needed - they are
 * temporary objects which are created when an SNMP request needs to
 * access to the instrumented MBean. 
 * The list of JmxMBeanAttrEntryImpl objects reflecting the attributes
 * of the mirrored MBean can also be temporarily cached in this object.
 *
 **/
public class JmxMBeanEntryImpl extends JmxMBeanEntry {

    // The ObjectName of the MBean mirrored by this entry.
    //
    private ObjectName name  = null;

    // The instantiated MBean mirrored by this entry, when this entry
    // was created from a remote SNMP SET request - see createFromRemote().
    //
    private transient Object     mbean = null;

    // Tell whether this entry is being created from remote.
    // This flag is initially set to false. If this entry is created from
    // a remote SET request, the createFromRemote() method will set it to
    // true. When the remote creation is eventually completed - when
    // the jmxMBeanRowStatus is switched to "active", this flag is finally
    // switched back to false. 
    //
    // This is necessary because some specific actions need to be completed
    // when the entry is created from a remote SNMP SET: in that case, a new
    // MBean needs to be created and registered in the MBeanServer - see
    // activate().
    //
    private boolean    remoteCreation = false;

    // Tell whether this entry is being deleted from remote. 
    //
    // This flag is initially set to true. If this entry is deleted from
    // a local MBeanServerNotification, the deleteFromRemote() method will 
    // set it to false, so that no attempt is made to unregister an already
    // unregistered MBean.
    // Otherwise, when the remote deletion is eventually completed - when
    // the jmxMBeanRowStatus is switched to "destroyed", and if this flag
    // is true, the MBean mirrored by this entry will be eventually
    // deregistered from the MBeanServer - see destroy().
    //
    private boolean    remoteDeletion = true;

    // List of attributes for the mirrors MBean that we want to expose 
    // through the jmxMBeanAttrTable. Serves as a temporary cache
    // for the rows of the JmxMBeanAttrTable that are linked to this row
    // of the jmxMBeanTable. The JmxMBeanAttrTable itself is virtual - so
    // the JmxMBeanAttrEntry objects are created on the fly when needed and
    // then temporarily cached into their JmxMBeanEntry. They are later
    // reclaimed by the garbage collector when the JmxMBeanEntry itself is
    // reclaimed (presumably at the end of the request, or when the weak
    // cache is garbage collected). 
    // See JmxMBeanTableMetaImpl.JmxMBeanTableCache
    //
    private JmxMBeanAttrEntryImpl[] attrList;

    // A pointer to the JmxMBeanServer group to which this entry belongs.
    //
    private final JmxMBeanTableMetaImpl myMeta;

    /**
     * Constructor for the "JmxMBeanEntry" conceptual row.
     */
    public JmxMBeanEntryImpl(SnmpMib myMib, JmxMBeanTableMetaImpl myMeta) {
	super(myMib);
	name     = null;
	attrList = null;

	// Initialize everything to null: the super class initialize
	// this variables with dummy values. Here we want to initialize
	// them with `null' so that we can later know whether the 
	// the corresponding variable has been properly set or is still
	// uninitialized (null means uninitialized).
	//
	JmxMBeanObjectName  = null;	
	JmxMBeanClassName   = null;
	JmxMBeanAttrCount   = null;
	JmxMBeanDescription = null;

	// By default, we will say that this row is notReady. As soon as
	// JmxMBeanObjectName and JmxMBeanClassName are set, we will switch
	// this value to "notInService".
	//
	JmxMBeanRowStatus   = new EnumJmxMBeanRowStatus("notReady");
	this.myMeta = myMeta;
    }

    /**
     * The ObjectName of the MBean mirrored by this entry.
     **/
    public synchronized ObjectName getObjectName() {
	return name;
    }


    /**
     * Setter for the "JmxMBeanAttrCount" variable.
     * If the row is being created from remote we will ignore this value.
     * Otherwise we must throw an exception.
     */
    public void setJmxMBeanAttrCount(Integer x) throws SnmpStatusException {
	checkJmxMBeanAttrCount(x);
    }

    /**
     * Checker for the "JmxMBeanAttrCount" variable.
     * If the row is being created from remote we will ignore this value.
     * Otherwise we must throw an exception.
     **/
    public synchronized void checkJmxMBeanAttrCount(Integer x) 
	throws SnmpStatusException {
	if (!remoteCreation)
	    throw new SnmpStatusException(SnmpStatusException.
					  snmpRspNotWritable);
    }

    /**
     * Setter for the "JmxMBeanDescription" variable.
     * If the row is being created from remote we will ignore this value.
     * Otherwise we must throw an exception.
     */
    public void setJmxMBeanDescription(String x) throws SnmpStatusException {
	checkJmxMBeanDescription(x);
    }

    /**
     * Checker for the "JmxMBeanDescription" variable.
     * If the row is being created from remote we will ignore this value.
     * Otherwise we must throw an exception.
     */
    public void checkJmxMBeanDescription(String x) throws SnmpStatusException {
	if (!remoteCreation)
	    throw new SnmpStatusException(SnmpStatusException.
					  snmpRspNotWritable);
    }

    /**
     * Setter for the "JmxMBeanClassName" variable.
     * Call checkJmxMBeanClassName(x) - this is useful in case several
     * values were supplied. 
     * If the row is already created  we must throw an exception.
     */
    public void setJmxMBeanClassName(String x) throws SnmpStatusException {
	checkJmxMBeanClassName(x);
	checkReady();
    }

    /**
     * Checker for the "JmxMBeanClassName" variable.
     * If the row is already created  we must throw an exception.
     */
    public synchronized void checkJmxMBeanClassName(String x) 
	throws SnmpStatusException {
	if (!remoteCreation)
	    throw new SnmpStatusException(SnmpStatusException.
					  snmpRspNotWritable);
	if (JmxMBeanClassName == null) {
	    if (x != null) {
		JmxMBeanClassName = 
		    JMX_MBEAN_SERVER_MIB_Impl.validJmxJavaClassNameTC(x);
		try {
		    final MBeanServer s = getMibServer();
		    mbean = s.instantiate(x);
		} catch (Exception ex) {
		    SnmpStatusException sn = 
			new SnmpStatusException(SnmpStatusException.
						snmpRspWrongValue);
		    sn.initCause(ex);
		    throw sn;
		}
	    }
	} else if (!JmxMBeanClassName.equals(JMX_MBEAN_SERVER_MIB_Impl.
					     validJmxJavaClassNameTC(x)))
	    throw new SnmpStatusException(SnmpStatusException.
					  snmpRspInconsistentValue);
    }

    /**
     * Setter for the "JmxMBeanObjectName" variable.
     * Call checkJmxMBeanObjectName(x) - this is useful in case several
     * values were supplied. 
     * If the row is already created  we must throw an exception.
     */
    public synchronized void setJmxMBeanObjectName(String x) 
	throws SnmpStatusException {
        if ((remoteCreation == false) && (JmxMBeanObjectName == null))  
	    initJmxMBean(x);
	else  checkJmxMBeanObjectName(x);
	checkReady();
    }

    /**
     * If both objectName and className are set and if the status is 
     * notReady we can switch it to notInService.
     **/
    private void checkReady() {
	if ((JmxMBeanRowStatus.intValue() == EnumRowStatus.notReady) 
	    && (JmxMBeanObjectName != null) 
	    && (JmxMBeanClassName  != null)) {
	    JmxMBeanRowStatus = 
		new EnumJmxMBeanRowStatus(EnumRowStatus.notInService);
	}
    }

    /**
     * Checker for the "JmxMBeanObjectName" variable.
     * If the row is already created  we must throw an exception.
     */
    public synchronized void checkJmxMBeanObjectName(String x) 
	throws SnmpStatusException {
	if (!remoteCreation)
	    throw new SnmpStatusException(SnmpStatusException.
					  snmpRspNotWritable);
	if (JmxMBeanObjectName == null) {
	    if (x != null) {
		JmxMBeanObjectName = 
		    JMX_MBEAN_SERVER_MIB_Impl.validJmxMBeanObjectNameTC(x);
		try {
		    final MBeanServer s = getMibServer();
		    name = new ObjectName(x);
		    if (s.isRegistered(name)) 
			throw new InstanceAlreadyExistsException();
		} catch (Exception ex) {
		    SnmpStatusException sn = 
			new SnmpStatusException(SnmpStatusException.
						snmpRspWrongValue);
		    sn.initCause(ex);
		    throw sn;
		}
	    } else if (!JmxMBeanObjectName.equals(JMX_MBEAN_SERVER_MIB_Impl.
						validJmxMBeanObjectNameTC(x)))
		throw new SnmpStatusException(SnmpStatusException.
					      snmpRspInconsistentValue);
	}
    }

    /**
     * Called by JmxMBeanServerImpl when the row is created from
     * remote. Check consistency of the supplied varbind list.
     **/
    public synchronized void createFromRemote(SnmpMibSubRequest req, 
					      SnmpVarBind rowStatusVb) 
	throws SnmpStatusException {

	remoteCreation = true;

	EnumRowStatus rowStatus = 
	    new EnumRowStatus((SnmpInt)(rowStatusVb.getSnmpValue()));

	final Enumeration e = req.getElements();

	SnmpOid     objectNameOid = new SnmpOid("jmxMBeanObjectName."+
						JmxMBeanIndex);
	SnmpOid     classNameOid = new SnmpOid("jmxMBeanClassName."+
					       JmxMBeanIndex);
	String objectName = null, className  = null;
	int    objectNameIndex=-1, classNameIndex=-1;

	for (;e.hasMoreElements();) {
	    SnmpVarBind vb = (SnmpVarBind) e.nextElement();
	    if (vb.getOid().equals(objectNameOid)) {
		if (vb.getSnmpValue() instanceof SnmpString) {
		    objectName = vb.getSnmpValue().toString();
		    objectNameIndex = req.getVarIndex(vb);
		} else throw new SnmpStatusException(SnmpStatusException.
						     snmpRspWrongType,
						     req.getVarIndex(vb));
	    } else if (vb.getOid().equals(classNameOid)) {
		if (vb.getSnmpValue() instanceof SnmpString) {
		    className  = vb.getSnmpValue().toString();
		    classNameIndex = req.getVarIndex(vb);
		} else throw new SnmpStatusException(SnmpStatusException.
						     snmpRspWrongType,
						     req.getVarIndex(vb));
	    }
	}

	if ((rowStatus.intValue() == EnumRowStatus.createAndGo)) {
	    if ((objectName == null) ||  (className == null)) {
		SnmpStatusException sn = 
		    new SnmpStatusException(SnmpStatusException.
					    snmpRspInconsistentValue);
		req.registerSetException(rowStatusVb,sn);
	    }
	}   

	try {
	    checkJmxMBeanObjectName(objectName);
	} catch (SnmpStatusException s) {
	    if (Boolean.getBoolean("info"))
		System.out.println("Bad ObjectName provided: " + objectName);
	    if (Boolean.getBoolean("debug")) s.printStackTrace();
	    throw new SnmpStatusException(s,objectNameIndex);
	}

	try {   
	    checkJmxMBeanClassName(className);
	} catch (SnmpStatusException s) {
	    if (Boolean.getBoolean("info"))
		System.out.println("Bad ClassName provided: " + className);
	    if (Boolean.getBoolean("debug")) s.printStackTrace();
	    throw new SnmpStatusException(s,classNameIndex);
	}
    }

    /**
     * Called by JmxMBeanServerImpl when the row is deleted locally.
     **/
    public synchronized void removeFromRemote(boolean remote) {
	remoteDeletion = remote;
    }

    /** 
     * Build the list of attributes. Does a lazy evaluation of the
     * attrList variable.
     **/
    synchronized JmxMBeanAttrEntryImpl[] getAttributes() {
	if (attrList != null) return attrList;
	ArrayList tmpAttrList = new ArrayList();
	int added = 0;
	int total = 0;
	try {
	    final MBeanServer s = getMibServer();
	    final MBeanInfo mbi = s.getMBeanInfo(name);
	    final MBeanAttributeInfo attrs[] = mbi.getAttributes();
	    total = attrs.length;
	    final int entryIndex = JmxMBeanIndex.intValue();
	    for (int i=0; i<attrs.length;i++) {
		final MBeanAttributeInfo attr = attrs[i];
		if (attr == null) continue;
		final JmxMBeanAttrEntryImpl e = 
		    new JmxMBeanAttrEntryImpl(myMeta.myMib,this);
		e.JmxMBeanAttrIndex = new Integer(added+1);
		e.JmxMBeanIndex     = JmxMBeanIndex;
		e.JmxMBeanAttrName  = attr.getName();
		e.JmxMBeanAttrType  = JMX_MBEAN_SERVER_MIB_Impl.
		    validJmxJavaClassNameTC(attr.getType());
		e.JmxMBeanAttrDescr = JMX_MBEAN_SERVER_MIB_Impl.
		    validDisplayStringTC(attr.getDescription());
		tmpAttrList.add(e);
		added++;
	    }
	    attrList = new JmxMBeanAttrEntryImpl[added];
	    tmpAttrList.toArray(attrList);
	} catch (Exception x) {
	    System.err.println("Failed to create all attributes for: " + name);
	    if (Boolean.getBoolean("debug")) x.printStackTrace();
	}
	if (Boolean.getBoolean("info")) 
	    System.out.println("Created " + added + "/" + total
			       + " attribute(s).");
	return attrList;
    }

    /**
     * Returns the attribute at the given index. 
     * This has the side effect of computing and caching all the rows of
     * jmxMBeanAttrTable that are linked to this object.
     **/
    JmxMBeanAttrEntryImpl getAttribute(int index) 
	throws SnmpStatusException {
	final JmxMBeanAttrEntryImpl[] attrs = getAttributes();
	if (attrs == null || index < 1 || index > attrs.length ) 
	    throw new SnmpStatusException(SnmpStatusException.noSuchInstance);
	return attrs[index-1];
    }

    /**
     * Change the row status. Eventually called by the runtime when
     * the row is created from remote. Called explicitly by 
     * JmxMBeanTableMetaImpl when the row is created on the fly.
     **/
    public synchronized void setJmxMBeanRowStatus(EnumJmxMBeanRowStatus x) 
	throws SnmpStatusException {
	switch (x.intValue()) {
	case EnumRowStatus.active:
	    if (! (JmxMBeanRowStatus.intValue() == EnumRowStatus.active))
		activate();
	    break;
	case EnumRowStatus.notReady:
	    break;
	case EnumRowStatus.notInService:
	    super.setJmxMBeanRowStatus(x);
	    break;
	case EnumRowStatus.destroy:
	    destroy();
	    break;
	default:
	    throw new SnmpStatusException(SnmpStatusException.
					  snmpRspInconsistentValue);
	}
    }

    /**
     * Activate the row. If the row was created from remote, this will
     * trigger the registration of the new MBean.
     **/
    private void activate() throws SnmpStatusException {
	if ((remoteCreation) && (mbean == null || name == null))
	    throw new SnmpStatusException(SnmpStatusException.snmpRspGenErr);
	try {

	    if (remoteCreation) {
		myMeta.registerMBean(mbean,name,this);
		initJmxMBean(name);
		remoteCreation = false;
		mbean = null;
	    } 

	    JmxMBeanRowStatus = 
		new EnumJmxMBeanRowStatus(EnumRowStatus.active);
	} catch (Exception x) {
	    SnmpStatusException sn = 
		new SnmpStatusException(SnmpStatusException.snmpRspGenErr);
	    sn.initCause(x);
	    throw sn;
	}
    }

    /**
     * Unregister the MBean mirrored by this row. Can only be called
     * from remote.
     **/
    private void destroy() throws SnmpStatusException {
	try {
	    JmxMBeanRowStatus = 
		new EnumJmxMBeanRowStatus(EnumRowStatus.notInService);
	    if (name != null && remoteDeletion)
		myMeta.unregisterMBean(name,this);
	    remoteCreation = false;
	    mbean = null; 
	    attrList = null;
	} catch (Exception x) {
	    SnmpStatusException sn = 
		new SnmpStatusException(SnmpStatusException.snmpRspGenErr);
	    sn.initCause(x);
	    throw sn;
	}
    }

    /**
     * Called when this row is created on the fly by JmxMBeanTableMetaImpl.
     **/
    private void initJmxMBean(String x) throws SnmpStatusException {
 	try {
	    name = new ObjectName(x);
	    JmxMBeanObjectName  = 
		JMX_MBEAN_SERVER_MIB_Impl.validJmxMBeanObjectNameTC(x);
	} catch (Exception ex) {
	    if (Boolean.getBoolean("debug")) ex.printStackTrace();
	    SnmpStatusException se = 
		new SnmpStatusException(SnmpStatusException.snmpRspWrongValue);
	    se.initCause(ex);
	    throw se;
	}
	initJmxMBean(name);
    }


    /**
     * Called when this row is created on the fly by JmxMBeanTableMetaImpl.
     **/
    private void initJmxMBean(ObjectName n) throws SnmpStatusException {
 	try {
	    final MBeanServer s = getMibServer();
	    final MBeanInfo   i = s.getMBeanInfo(name);
	    JmxMBeanClassName = JMX_MBEAN_SERVER_MIB_Impl.
		validJmxJavaClassNameTC(i.getClassName());
	    JmxMBeanAttrCount = new Integer(i.getAttributes().length);
	    JmxMBeanDescription = JMX_MBEAN_SERVER_MIB_Impl.
		validDisplayStringTC(i.getDescription());
	} catch (Exception ex) {
	    if (Boolean.getBoolean("debug")) ex.printStackTrace();
	    SnmpStatusException se = 
		new SnmpStatusException(SnmpStatusException.snmpRspWrongValue);
	    se.initCause(ex);
	    throw se;
	}
    }

    /**
     * The MBeanServer that is mirrored by this MIB. 
     * Note: the MIB is not expected to be registered in an MBeanServer.
     *       It is only registered in the SnmpAdaptorServer.
     **/
    public MBeanServer getMibServer() {
	return  myMeta.getMibServer();
    }
    
}
