/*
 * @(#)file      StandAloneAgent.java
 * @(#)author    Sun Microsystems, Inc.
 * @(#)version   1.17
 * @(#)lastedit  04/04/07
 *
 * Copyright 2004 Sun Microsystems, Inc. All rights reserved.
 * SUN PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

// Java imports
//
import java.net.InetAddress;
import java.net.UnknownHostException;

// JDMK imports
//
import com.sun.management.snmp.SnmpVarBindList;
import com.sun.management.snmp.SnmpVarBind;
import com.sun.management.snmp.SnmpOid;
import com.sun.management.snmp.SnmpString;
import com.sun.management.snmp.SnmpNull;
import com.sun.management.snmp.SnmpInt;
import com.sun.management.snmp.SnmpDefinitions;
import com.sun.management.snmp.SnmpIpAddress;
import com.sun.management.snmp.SnmpCounter64;
import com.sun.management.snmp.SnmpEngine;

import com.sun.management.comm.SnmpV3AdaptorServer;
import com.sun.management.comm.SnmpAdaptorServer;
import com.sun.management.snmp.agent.SnmpTrap;
import com.sun.management.snmp.agent.SnmpProxy;
import com.sun.management.internal.snmp.SnmpEngineImpl;

/**
 * The StandAloneAgent class provides a simple example on how to use
 * the SNMP protocol adaptor.
 * <p>
 * A MIB implementing a fake System Group in the OID space of MIB-II
 * is instrumented by this agent. The MIB is loaded and initialized.
 * As such you can now see the MIB using your favorite SNMP manager.
 * <p>
 * When calling the program, you must specify:
 * <ul>
 *     <li>port: port number to use.</li>
 * </ul>
 * <p>
 * In this example, the SNMP adaptor is started on the supplied port,
 * and the traps are sent to the port 'port + 1'.
 */
public class StandAloneAgent {

    // The SNMP Adaptor.
    //
    static SnmpAdaptorServer snmpAdaptor = null;

    /**
     * This class is used to generate traps.
     **/
    private static class TrapGenerator  extends Thread {

        SnmpV3AdaptorServer server = null;
        SnmpVarBindList lgarbage = null;
    
        private void sendV2() {

            println("V2 TRAP to send\n");
        
            //Send a V2 trap.
            try {
                server.snmpV2Trap(SnmpTrap.egpNeighborLossOid, lgarbage);
                println("Trap V2 sent!");  
            } catch (Exception e) { 
                e.printStackTrace(); 
            }
        }
        
        private void sendV1() {
            
            println("V1 TRAP to send\n");
        
            // Send a V1 trap.
            try{
                server.snmpV1Trap(0,0,lgarbage);
                println("Trap V1 sent!");
            } catch (Exception e) {
                e.printStackTrace(); 
            }
        }

        public void run() {
            int c = 0;
            while (true) {
                if (c != '\n')
                    println("\nType 1 in order to launch snmp V1 traps," +
                            " 2 for snmp V2 traps.");
                try {
                    c =  java.lang.System.in.read();
                    if (c == '\n') continue;
                } catch (Exception e) {
		}

                if (c == '1') { sendV1(); continue; }
                if (c == '2') { sendV2(); continue; }
                println("(1 or 2). Ctrl+C if you want to stop.");
            }
        }

        public TrapGenerator(SnmpAdaptorServer server) {
        
            this.server = (SnmpV3AdaptorServer) server;
        
            //Initialize a list of Varbinds
            lgarbage = new SnmpVarBindList();
            lgarbage.addVarBind(new 
                SnmpVarBind(new SnmpOid("1.2.3.4.5.6.0.0"), 
                            new SnmpString("Test values")));
            lgarbage.addVarBind(new 
                SnmpVarBind(new SnmpOid("1.2.3.4.5.6.1.0"), 
                            new SnmpNull()));
            lgarbage.addVarBind(new 
                SnmpVarBind(new SnmpOid("1.2.3.4.5.6.2.0"), 
                            new SnmpInt (43)));
            lgarbage.addVarBind(new 
                SnmpVarBind(new SnmpOid("1.2.3.4.5.6.3.0"), 
                            new SnmpString ("Test values")));

            //Instantiate an address OID varBind
            SnmpVarBind vb = null;
            try {
                final byte[] inet = 
                    InetAddress.getLocalHost().getAddress();
                final SnmpIpAddress addr = new SnmpIpAddress(inet);
                vb = new SnmpVarBind(SnmpTrap.snmpTrapAddressOid,addr);
            } catch (UnknownHostException e) {
                vb = new SnmpVarBind(SnmpTrap.snmpTrapAddressOid, 
                                     new SnmpIpAddress("123.33.33.33")); 
            }
            lgarbage.addVarBind(vb);
        }
    }

    /**
     * Main entry point.
     */
    public static void main(String argv[]) {
        try {
            //
            // SNMP specific code:
            //
            if (argv.length != 1) {
                usage();
                java.lang.System.exit(1);
            }
            
            // Access udp port
            int snmpPort = Integer.parseInt(argv[0]);
            
            println("NOTE: The SNMP Adaptor is bound to " +
                    "UDP port " + snmpPort);
            println("NOTE: The SNMP Adaptor sends traps to " +
                    "UDP port " + (snmpPort + 1));
            snmpAdaptor = new SnmpV3AdaptorServer(snmpPort);
            snmpAdaptor.setTrapPort(new Integer(snmpPort + 1));

            // The USM MIB is registered in the adaptor.
            ((SnmpV3AdaptorServer)snmpAdaptor).registerUsmMib();

            // Set requests are enabled for V1 and V2 managers
            ((SnmpV3AdaptorServer)snmpAdaptor).enableSnmpV1V2SetRequest();

            snmpAdaptor.start();

            // Instantiate a Mib
            MIB1_MIB mib1 = new MIB1_MIB();
            mib1.init();
	    mib1.setSnmpAdaptor(snmpAdaptor);

            println("Press <Return> if you want to send traps");
            java.lang.System.in.read();
            //Instantiate and start the TrapGenerator.
            TrapGenerator trapGenerator = new TrapGenerator(snmpAdaptor);
            trapGenerator.start();
            println("Press <CTRL+C> in order to kill the agent");
            trapGenerator.join();
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    public static void  usage() {
        println("java -Djdmk.security.file=jdmk.security " +
		"StandAloneAgent <subagent_port>");
        println("where");
        println("\t-subagent_port: port number to use.");
    }

    /**
     * print/println stuff...
     */

    private final static void println(String msg) {
        java.lang.System.out.println(msg);
    }

    private final static void print(String msg) {
        java.lang.System.out.print(msg);
    }    
}
