/*
 * @(#)file      Manager.java
 * @(#)author    Sun Microsystems, Inc.
 * @(#)version   1.32
 * @(#)lastedit  04/04/21
 *
 * Copyright 2004 Sun Microsystems, Inc. All rights reserved.
 * SUN PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

// Java imports
//
import java.io.IOException;

// JDMK imports
//
import com.sun.management.snmp.SnmpDefinitions;
import com.sun.management.snmp.SnmpOid;
import com.sun.management.snmp.SnmpInt;
import com.sun.management.snmp.SnmpString;
import com.sun.management.snmp.SnmpVarBind;
import com.sun.management.snmp.SnmpVarBindList;
import com.sun.management.snmp.manager.SnmpPeer;
import com.sun.management.snmp.manager.SnmpParameters;
import com.sun.management.snmp.manager.SnmpRequest;
import com.sun.management.snmp.manager.SnmpSession;

import com.sun.management.snmp.SnmpOidDatabaseSupport;

/**
 * The Manager class is a simple example showing how to use the SNMP
 * Manager API.
 *
 * When calling the program, you must specify:
 *      - host: hostname of the SNMP agent you want to query.
 *      - port: port number to use.
 *
 * The provided SNMP agent uses port 8085.
 * To query this agent, invoke the program as follows:
 *      java Manager <hostname> 8085
 */

public class Manager {

    /**
     * The main method allows you to specify the following optional
     * command-line parameters:
     *  1) host: hostname of the SNMP agent you want to query.
     *  2) port: port number to use.
     *
     * If you wish to query the provided SNMP agent, use port 8085.
     */
    public static void main(String argv[]) {

        if (argv.length != 2) {
            usage();
            java.lang.System.exit(1);
        }

        String host = argv[0];
        String port = argv[1];

        try {

            // Initialize the SNMP Manager API.
            // Specify the OidTables containing all the MIB Demo and
	    // MIB II knowledge. Use the OidTables generated by mibgen
	    // when compiling MIB Demo and MIB II.
            //
            SnmpOidDatabaseSupport oidDB = new SnmpOidDatabaseSupport();
            SnmpOid.setSnmpOidTable(oidDB);
            oidDB.add(new RFC1213_MIBOidTable());
            oidDB.add(new DEMO_MIBOidTable());

            SnmpPeer agent = new SnmpPeer(host, Integer.parseInt(port));
            SnmpParameters params = new SnmpParameters("public", "private");
            SnmpSession session = new SnmpSession("Manager session");

            // We update the timeout to let the agent enough time
            // to do his job before retry.
            //
            agent.setTimeout(100000);
            agent.setParams(params);
            session.setDefaultPeer(agent);

            // We query demo and system groups of DEMO_MIB and RFC1213_MIB
	    // respectively.
            // The DEMO_MIB has been loaded by the local Agent.
            // The RFC1213_MIB has been loaded by the remote StandAloneAgent.
            //
            SnmpRequest request= null;
            SnmpVarBindList result= null;
            boolean completed= false;
            int errorStatus= 0;
            int errorIndex= 0;

            SnmpVarBindList list= new SnmpVarBindList("Manager varbind list");

            SnmpVarBind avar1 = new SnmpVarBind("demoString.0");
            SnmpVarBind avar2 = new SnmpVarBind("demoInteger.0");
            SnmpVarBind avar3 = new SnmpVarBind("demoOid.0");
            SnmpVarBind avar4 = new SnmpVarBind("sysDescr.0");
            SnmpVarBind avar5 = new SnmpVarBind("sysContact.0");
            SnmpVarBind avar6 = new SnmpVarBind("sysName.0");

            list.addVarBind(avar1);
            //list.addVarBind(avar2);	// RO: the set operation would fail...
            list.addVarBind(avar3);
            //list.addVarBind(avar4);	// RO: the set operation would fail...
            list.addVarBind(avar5);
            list.addVarBind(avar6);

            // First get request.
            //
            java.lang.System.out.println("\n>> ==============================" +
					 "===================================" +
					 "====");
            java.lang.System.out.println(">> Manager: Sending get request to " +
					 "SNMP agent on " + host +
					 " using port " + port);
            request = session.snmpGetRequest(null, list);
            java.lang.System.out.println("\nRequest:\n" + request.toString());

            completed = request.waitForCompletion(10000);
            if (completed == false) {
                java.lang.System.out.println("\nRequest timed out: check " +
					     "reachability of agent.");
                java.lang.System.exit(1);
            }
            errorStatus = request.getErrorStatus();
            errorIndex = request.getErrorIndex() + 1;
            if (errorStatus != SnmpDefinitions.snmpRspNoError) {
                java.lang.System.out.println("\nError status = " +
				 SnmpRequest.snmpErrorToString(errorStatus));
                java.lang.System.out.println("Error index = " + errorIndex);
            } else {
                result = request.getResponseVarBindList();
                java.lang.System.out.println("\nResult: \n" +
					     result.varBindListToString());
            }

            // Set request.
            //
            avar1.setSnmpValue(new SnmpString("! My new demoString !"));
            avar2.setSnmpValue(new SnmpInt(456));
            avar3.setSnmpValue(new SnmpOid("1.3.6.1.4.1.43"));
            avar4.setSnmpValue(new SnmpString("! My new sysDescr !"));
            avar5.setSnmpValue(new SnmpString("! My new sysContact !"));
            avar6.setSnmpValue(new SnmpString("! My new sysName !"));
            java.lang.System.out.println("\n>> ==============================" +
					 "===================================" +
					 "====");
            java.lang.System.out.println(">> Manager: Sending set request to " +
					 "SNMP agent on " + host +
					 " using port " + port);
            request = session.snmpSetRequest(null, list);
            java.lang.System.out.println("\nRequest:\n" + request.toString());

            completed = request.waitForCompletion(10000);
            if (completed == false) {
                java.lang.System.out.println("\nRequest timed out: check " +
					     "reachability of agent.");
                java.lang.System.exit(1);
            }
            errorStatus = request.getErrorStatus();
            errorIndex = request.getErrorIndex() + 1;
            if (errorStatus != SnmpDefinitions.snmpRspNoError) {
                java.lang.System.out.println("\nError status = " +
				 SnmpRequest.snmpErrorToString(errorStatus));
                java.lang.System.out.println("Error index = " + errorIndex);
            } else {
                result = request.getResponseVarBindList();
                java.lang.System.out.println("\nResult: \n" +
					     result.varBindListToString());
            }

            // Second get request.
            //
            java.lang.System.out.println("\n>> ==============================" +
					 "===================================" +
					 "====");
            java.lang.System.out.println(">> Manager: Sending get request to " +
					 "SNMP agent on " + host +
					 " using port " + port);
            request = session.snmpGetRequest(null, list);
            java.lang.System.out.println("\nRequest:\n" + request.toString());

            completed = request.waitForCompletion(10000);
            if (completed == false) {
                java.lang.System.out.println("\nRequest timed out: check " +
					     "reachability of agent.");
                java.lang.System.exit(1);
            }
            errorStatus = request.getErrorStatus();
            errorIndex = request.getErrorIndex() + 1;
            if (errorStatus != SnmpDefinitions.snmpRspNoError) {
                java.lang.System.out.println("\nError status = " +
				 SnmpRequest.snmpErrorToString(errorStatus));
                java.lang.System.out.println("Error index = " + errorIndex);
            } else {
                result = request.getResponseVarBindList();
                java.lang.System.out.println("\nResult: \n" +
					     result.varBindListToString());
            }

            // getNext request.
            //
            java.lang.System.out.println("\n>> ==============================" +
					 "===================================" +
					 "====");
            java.lang.System.out.println(">> Manager: Sending getNext request" +
					 " to SNMP agent on " + host +
					 " using port " + port);
            request = session.snmpGetNextRequest(null, list);
            java.lang.System.out.println("\nRequest:\n" + request.toString());

            completed = request.waitForCompletion(100000);
            if (completed == false) {
                java.lang.System.out.println("\nRequest timed out: check " +
					     "reachability of agent.");
                java.lang.System.exit(1);
            }
            errorStatus = request.getErrorStatus();
            errorIndex = request.getErrorIndex() + 1;
            if (errorStatus != SnmpDefinitions.snmpRspNoError) {
                java.lang.System.out.println("\nError status = " +
				 SnmpRequest.snmpErrorToString(errorStatus));
                java.lang.System.out.println("Error index = " + errorIndex);
            } else {
                result = request.getResponseVarBindList();
                java.lang.System.out.println("\nResult: \n" +
					     result.varBindListToString());
            }

            // Nicely stop the session.
            //
            session.destroySession();
            java.lang.System.exit(0);

        } catch(Exception e) {
            java.lang.System.err.println("Exception occurred: " + e);
            e.printStackTrace();
            java.lang.System.exit(1);
        }
    }

    public static void  usage() {
        java.lang.System.out.println("\nUsage: java Manager <host> <port>");
        java.lang.System.out.println("where");
        java.lang.System.out.println("\t-host: hostname of the SNMP agent " +
				     "you want to query.");
        java.lang.System.out.println("\t-port: port number to use.\n");
    }
}
