/*
 * @(#)file      CreateUsmMibUser.java
 * @(#)author    Sun Microsystems, Inc.
 * @(#)version   1.15
 * @(#)lastedit  04/04/21
 *
 * Copyright 2004 Sun Microsystems, Inc. All rights reserved.
 * SUN PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

// Java imports
//
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.util.HashMap;

// JDMK imports
//
import com.sun.management.snmp.SnmpDefinitions;
import com.sun.management.snmp.SnmpEngine;
import com.sun.management.snmp.SnmpEngineId;
import com.sun.management.snmp.SnmpEngineParameters;
import com.sun.management.snmp.SnmpInt;
import com.sun.management.snmp.SnmpOid;
import com.sun.management.snmp.SnmpStatusException;
import com.sun.management.snmp.SnmpString;
import com.sun.management.snmp.SnmpUsmKeyHandler;
import com.sun.management.snmp.SnmpVarBind;
import com.sun.management.snmp.SnmpVarBindList;
import com.sun.management.snmp.manager.SnmpEventReportDispatcher;
import com.sun.management.snmp.manager.SnmpParameters;
import com.sun.management.snmp.manager.SnmpParams;
import com.sun.management.snmp.manager.SnmpPeer;
import com.sun.management.snmp.manager.SnmpRequest;
import com.sun.management.snmp.manager.SnmpSession;
import com.sun.management.snmp.manager.usm.SnmpUsmParameters;
import com.sun.management.snmp.manager.usm.SnmpUsmPeer;

import com.sun.management.internal.snmp.SnmpTools;
import com.sun.management.snmp.EnumRowStatus;
import com.sun.management.snmp.SnmpOidTableSupport;

/**
 * The class is a tool example showing how to use the SNMP V3 Manager 
 * API in order to instantiate new user in an agent USM MIB.
 *
 * <p>When calling the program, you must specify:
 * <br>     - user name: The user name to use when making requests.
 * <br>     - security level: The security level to use when making requests.
 * <br>     - host: hostname of the SNMP agent you want to query.
 * <br>     - port: port number to use.
 * 
 * <p>As a reminder, if you wish to query the SNMP agent example provided
 *    as part of Java DMK, use port 8085.
 *
 * <p> A new Java DMK java property is used to set a security file: 
 * <pre>
 *     -Djdmk.security.file=<i>file</i>
 * </pre>
 * <p>It contains security information related to USM users. Have a look 
 *    at the READMEV3 file in order to find out how SNMP V3 managers are 
 *    configured.
 */

public class CreateUsmMibUser {

    // Some internal variable used to acquire data.
    //
    String host = null; 
    String port = null;
    BufferedReader in = null;
    String engId = "0x000000000000000000000002";
    SnmpEngineId engineId = null;
    String userName = "myNewUser";
    String cloneFromUser = "defaultUser";
    String securityLevel = "authPriv";
    String authProtocol = "usmHMACMD5AuthProtocol";
    String authRandom = "00000000000000000000000000000000";
    String authPasswd = "maplesyrup";
    String authNewPasswd = "newsyrup";
    String privRandom = "00000000000000000000000000000000";
    String privPasswd = "maplesyrup";
    String privNewPasswd = "newsyrup";
    SnmpSession session = null;
    SnmpUsmPeer agent = null;
    SnmpEngine engine = null;
    int step = 0;
    boolean undo = true;
    SnmpUsmKeyHandler handler = null;
    HashMap users = new HashMap();
    
     /**
     * Main entry point.
     * <p>When calling the program, you must specify:
     * <br>     - user name: The user name to use when making requests.
     * <br>     - security level: The security level to use when making 
     *            requests.
     * <br>     - host: hostname of the SNMP agent you want to query.
     * <br>     - port: port number to use.
     *
     * <p> As a reminder, if you wish to query the SNMP agent example
     *     provided as part of Java DMK, use port 8085.
     */
    public static void main(String argv[]) {
        if (argv.length != 4) {
            usage();
            java.lang.System.exit(1);
        }

        // Check arguments first
        //
        final String name = argv[0];
        final String securityLevel = argv[1];
        final String host = argv[2];
        final String port = argv[3];

        // Create the creator.
        //
        CreateUsmMibUser creator = 
            new CreateUsmMibUser(host, port, name, securityLevel);
        
        //The main loop
        //
        while (true) {
            println("\n\n***** NEW USER INPUTS :");
            //Acquire data
            creator.readArgs();
            //Create the user.
            creator.createUser();
        }
        
    }


    /**
     * Constructor. Initializes the remote connection and various IO.
     */
    public CreateUsmMibUser(String host, 
                            String port,
                            String name,
                            String securityLevel) {
        println("\nInitializing creator.");
        this.host = host;
        this.port = port;
        
        //Initialize engine ID with default value
        try {
            engineId = SnmpEngineId.createEngineId(engId);
        } catch (Exception e) {
            println(e.getMessage());
            java.lang.System.exit(0);
        }

        // Read from stdin
        //
        InputStreamReader i = new InputStreamReader(java.lang.System.in);
        in = new BufferedReader(i);

        final SnmpEngineParameters parameters = new SnmpEngineParameters();
        parameters.activateEncryption();

        // Create the SNMP Manager Session
        //
        try {
            session= new SnmpSession(parameters, null,
                                     "CreateUsmMibUser session", null);
        } catch (SnmpStatusException e) {
            println(e.getMessage());
            java.lang.System.exit(0);
        } catch (IllegalArgumentException e) {
            println(e.getMessage());
            java.lang.System.exit(0);
        }

        try {
            // get the session the engine.
            //
            engine = session.getEngine();

            // *********** INTERESTING STEP ****************
            // VERY USEFULL CLASS. The Usm Key handler.
            // IT ALLOWS YOU TO DO KEY CHANGES.
            //
            // This corresponds to the key change algorithm described
            // in RFC 2574.
            //
            handler = engine.getUsmKeyHandler();

            // Create the peer to the agent where the users will be created.
            //
            agent = new SnmpUsmPeer(engine,host,Integer.parseInt(port));
            boolean discoveryFlag = false;
            
            // Security parameters when sending set requests.
            //
            final SnmpUsmParameters p = new SnmpUsmParameters(engine, name);
            
            // Analyze the security level
            // If the security level is authPriv or authNoPriv, we will
            // need to send a "discovery" request in order to discover
            // the value of the agent's engineBoots and engineTime.
            //
            if (securityLevel.equals("authPriv")) {
                discoveryFlag = true;
                p.setSecurityLevel(SnmpDefinitions.authPriv);
            }
            if (securityLevel.equals("authNoPriv")) {
                discoveryFlag = true;
                p.setSecurityLevel(SnmpDefinitions.authNoPriv);
            }
            if (securityLevel.equals("noAuthNoPriv")) {
                p.setSecurityLevel(SnmpDefinitions.noAuthNoPriv);
            }
            
            p.setContextEngineId(agent.getEngineId().getBytes());
            
            agent.setParams(p);
            
            // Sends the discovery request if necessary.
            //
                if (discoveryFlag)
                    agent.processUsmTimelinessDiscovery();

        } catch (Exception e) {
            println("Unable to initialize creator." +
                    " Check whether the agent is running" +
                    " and/or whether your USM configuration is OK");
            java.lang.System.exit(0);
        }
        session.setDefaultPeer(agent);
    }
    
    /**
     * The readArgs() method is not very interesting in the context of 
     * this example. It is a simple input line parsing.
     */
    public void readArgs() {
        try {
            boolean ko = true;
            while (true) {
                ko = true;
                while (ko) {
                    try {
                        if (engineId == null ) {
                            print("\nType the engine Id :");
                            engId = in.readLine();
                            engineId = SnmpEngineId.createEngineId(engId);
                        } else {
                            print("\nType the engine Id" +
                                  " (press return to accept) ["+
                                  engineId.toString() +"]:");
                            engId = in.readLine();
                            if (!engId.equals("")) 
                                engineId = 
                                    SnmpEngineId.createEngineId(engId);
                        }
                        ko = false;
                    } catch (Exception e) {
                        println("Invalid engine Id!");
                    }
                }
                
                ko = true;
                while (ko) {
                    if (userName == null) {
                        print("\nType the new user name :");
                        userName = in.readLine();
                    } else {
                        print("\nType the new user name"+ 
                              " (press return to accept) [" + userName +"]:");
                        String tmpUser = in.readLine();
                        if (!tmpUser.equals(""))
                            userName = tmpUser;
                    }
                    if (users.get(userName) != null) {
                        print("\n Already created user. Change user name!");
                    } else {
                        users.put(userName, userName);
                        ko = false;
                    }
                }
                
                if (cloneFromUser == null ) {
                    print("\nType the clone from user name :");
                    cloneFromUser = in.readLine();
                } else {
                    print("\nType the clone from user name"+ 
                          " (press return to accept) [" + cloneFromUser +"]:");
                    String tmpClone = in.readLine();
                    if (!tmpClone.equals(""))
                        cloneFromUser = tmpClone;
                }
                while (ko) {
                    if (securityLevel == null) {
                        print("\nType the security level :");
                        securityLevel = in.readLine();
                    } else {
                        print("\nType the security level" +
                              " (press return to accept) ["+
                              securityLevel +"]:");
                        String tmpSec = in.readLine();
                        if (!tmpSec.equals(""))
                            securityLevel = tmpSec;
                    }
                    if (securityLevel.equals("authPriv")) {
                        if (privPasswd == null) {
                            print("\nType the old priv password :");
                            privPasswd = in.readLine();
                        } else {
                            print("\nType the old priv password" + 
                                  " (press return to accept) ["+
                                  privPasswd+"]:");
                            String tmpPriv = in.readLine();
                            if (!tmpPriv.equals(""))
                                privPasswd = tmpPriv;
                        }
                        if (privNewPasswd == null) {
                            print("\nType the new priv password :");
                            privNewPasswd = in.readLine();
                        } else {
                            print("\nType the new priv password" + 
                                  " (press return to accept) ["+
                                  privNewPasswd+"]:");
                            String tmpPriv = in.readLine();
                            if (!tmpPriv.equals(""))
                                privNewPasswd = tmpPriv;
                        }
                        if (privRandom == null) {
                            print("\nType the priv key random :");
                            privRandom = in.readLine();
                        } else {
                            print("\nType the priv key random" +
                                  " (press return to accept) ["+
                                  privRandom+"]:");
                            String tmpRand = in.readLine();
                            if (!tmpRand.equals(""))
                                privRandom = tmpRand;
                        }
                        ko = false;
                    }
                    
                    if (securityLevel.equals("authNoPriv") || 
                       securityLevel.equals("authPriv")) {
                        if (authProtocol == null) {
                            print("\nType the auth protocol :");
                            authProtocol = in.readLine();
                        } else {
                            print("\nType the auth protocol"+
                                  " (press return to accept) ["+
                                  authProtocol+"]:");
                            String tmpProto = in.readLine();
                            if (!tmpProto.equals(""))
                                authProtocol = tmpProto;
                        }
                        if (authPasswd == null) {
                            print("\nType the old auth password :");
                            authPasswd = in.readLine();
                        } else {
                            print("\nType the old auth password" +
                                  " (ret to accept [" + authPasswd +"]:");
                            String tmpAuth = in.readLine();
                            if (!tmpAuth.equals(""))
                                authPasswd = tmpAuth;
                        }
                        if (authNewPasswd == null) {
                            print("\nType the new auth password :");
                            authNewPasswd = in.readLine();
                        } else {
                            print("\nType the new auth password" +
                                  " (ret to accept [" + authNewPasswd +"]:");
                            String tmpAuth = in.readLine();
                            if (!tmpAuth.equals(""))
                                authNewPasswd = tmpAuth;
                        }
                        if (authRandom == null) {
                            print("\nType the auth key random :");
                            authRandom = in.readLine();
                        } else {
                            print("\nType the auth key random" +
                                  " (press return to accept) ["+
                                  authRandom+"]:");
                            String tmpRand = in.readLine();
                            if (!tmpRand.equals(""))
                                authRandom = tmpRand;
                        }
                        ko = false;
                    }
                    
                    if (securityLevel.equals("noAuthNoPriv"))
                        ko = false;
                    if (ko) {
                        println("Invalid Security level" +
                                " (" + "\"noAuthNoPriv\"" + ", " +
                                "\"authNoPriv\"" + ", " + "\"authPriv\"" +
                                ")");
                        securityLevel = null;
                    }
                }
                
                println("********** Input summary ************ \n");
                println("\t* Engine Id : " + engineId.toString());
                println("\t* New user name : " + userName);
                println("\t* Clone from : " + cloneFromUser);
                println("\t* Security level : " + securityLevel);
                if (securityLevel.equals("authPriv")) {
                    println("\t* Old priv password : " + privPasswd);
                    println("\t* New priv password : " + privNewPasswd);
                    println("\t* Priv key random : " + privRandom);
                }
                
                if (securityLevel.equals("authNoPriv") || 
                    securityLevel.equals("authPriv")) {
                    println("\t* Auth protocol : " + authProtocol);
                    println("\t* Old auth password : " + authPasswd);
                    println("\t* New auth password : " + authNewPasswd);
                    println("\t* Auth key random : " + authRandom);
                }
                print("Do you agree (yes, no) [yes]:");
                String ans = in.readLine();
                if (ans.equals("yes") ||
                    ans.equals(""))
                    break;
                else
                    continue;
            }
        } catch (Exception e) {
            println(e.getMessage());
            java.lang.System.exit(0);
        }
    }

    /**
     * Regular Java DMK manager API SNMP Get and Set requests. 
     * Idem other SNMP examples.
     */
    private SnmpVarBindList doRequest(SnmpVarBindList list, boolean get) {
        try {
            SnmpRequest request = null;
            if (get) 
                request = session.snmpGetRequest(null, list);
            else
                request = session.snmpSetRequest(null, list);
            
            final boolean completed = request.waitForCompletion(10000);
            
            // Check for a timeout of the request.
            //
            if (completed == false) {
                throw new IllegalArgumentException("Timeout");
            }
            
            // Now we have a response. Check if the response contains 
            // an error.
            //
            final int errorStatus = request.getErrorStatus();
            if (errorStatus != SnmpDefinitions.snmpRspNoError) {
                throw new IllegalArgumentException(
                                 SnmpRequest.snmpErrorToString(errorStatus));
            }
            
            // Extract the returned VarBind List
            //
            SnmpVarBindList result = request.getResponseVarBindList();

            // Now we may display the content of the result.
            //
            // println("Result: \n" + result);
            
            return result;
            
        } catch (Exception e) {
            throw new IllegalArgumentException(e.getMessage());
        }
    }


    /**
     * THE MAIN INTERESTING CODE. IT SEQUENTIALY CREATES A NEW USER.
     */
    public void createUser() {
        try {
            // Compute the row index for the new user.
            // Make use of SnmpString and SnmpEngineId classes
            // * toOid() methods are called on both classes in order 
            //   to build the index.
            // The index is composed of the engineId, and the user name.
            //
            final SnmpOid userNameOid = new SnmpString(userName).toOid();
            final String userIndex = new String(engineId.toOid() + "." + 
                                          userName.length() + "." + 
                                          userNameOid.toString());

            // USM MIB provides a dedicated oid in order to guard the 
            // MIB against concurrent access. This is a spinlock.
            // Create the spinLock VarBindList
            //
            final SnmpVarBindList spinlist = 
                new SnmpVarBindList("SpinLock varbind list");

            // Spinlock OID.
            //
            final SnmpOid oidSpin = new SnmpOid("1.3.6.1.6.3.15.1.2.1.0");
            final SnmpVarBind spin = new SnmpVarBind(oidSpin);
            spinlist.addVarBind(spin);


            // OID of the RowStatus object controlling the user row 
            // creation.
            //
            final SnmpOid statusOid = 
                new SnmpOid("1.3.6.1.6.3.15.1.2.2.1.13" + "." + userIndex);


            // Each step of the creation process are flaged with an int 
            // step-value.
            // The int is reset at the end of the process. If a step fails,
            // the process starts from the failed point (after a complete
            // the user has re-entered the data).
            //

            // The row creation (step 0)
            //
            if (step == 0) {

                // First get the spinlick.
                //
                final SnmpVarBindList result = doRequest(spinlist, true);
               

                // Compute the cloneFrom oid.
                //
                final SnmpOid cloneOid  = 
                    new SnmpOid("1.3.6.1.6.3.15.1.2.2.1.4"+"."+userIndex);
                
                // Compute the cloneFrom value.
                // Quite complicated : 
                //   1) clone from user name to Oid.
                //   2) Concatenate securityName oid, engineId, clone 
                //      from user length, clone from user name oid.
                //
                // See RFC 2574 for more details.
                //
                final SnmpOid userCloneOid = 
                    new SnmpString(cloneFromUser).toOid();
                final SnmpOid cloneValue = 
                    new SnmpOid("1.3.6.1.6.3.15.1.2.2.1.3"+ "." +
                                engineId.toOid() + "." + 
                                cloneFromUser.length() + "." +
                                userCloneOid.toString());
                
                
                // Create the complete VarBindList for the user creation
                // request (spinlock, status, cloneFrom)
                //
                final SnmpVarBindList list = 
                    new SnmpVarBindList("Create varbind list");

                // Add the value of the spinlock previously retrieved
                // from the agent.
                // 
                list.addVarBind(result.getVarBindAt(0));

                // Add the status createAndWait
                //
                final SnmpInt status = 
                    new SnmpInt(EnumRowStatus.createAndWait);
                SnmpVarBind rowStatus = 
                    new SnmpVarBind(statusOid, status);
                list.addVarBind(rowStatus);

                // Add the cloneFromUser.
                //
                SnmpVarBind cloneFrom = 
                    new SnmpVarBind(cloneOid, cloneValue);
                list.addVarBind(cloneFrom);

                // Do the set request.
                //
                doRequest(list, false);
                
                println("***** New user [" + userName + "] created.");
                undo = false;
                step = 1;
            }


            // Various key changes according to security level.
            
            // Privacy key change (step 1)
            //
            if (step == 1) {
                // Privacy key change is needed
                //
                if (securityLevel.equals("authPriv")) {
                    println("***** Doing Priv key change");
                    
                    // Construct the VarBindList.
                    //
                    final SnmpVarBindList list = 
                        new SnmpVarBindList("Priv key change varbind list");

                    // Get the spinLock and add the result to VarBindList
                    //
                    final SnmpVarBindList result = doRequest(spinlist, true);
                    list.addVarBind(result.getVarBindAt(0));

                    // Handle the random part of the Priv key change.
                    // Make use of SnmpString.HexToChar method.
                    // This is a useful method that converts an 
                    // hexadecimal binary string in a String.
                    // Once converted, the String bytes contains the 
                    // binary value, not the ascii one.
                    //
                    final String hexRandom = 
                        SnmpString.HexToChar(privRandom);
                    final SnmpString random = 
                        new SnmpString(hexRandom.getBytes());
                    
                    // Compute the random Oid.
                    //
                    final SnmpOid randomOid = 
                        new SnmpOid("1.3.6.1.6.3.15.1.2.2.1.11"+"."+
                                    userIndex);

                    // Create the random varbind
                    //
                    final SnmpVarBind randomVar = 
                        new SnmpVarBind(randomOid, random);
                    
                    // Add the random in the VarBindList
                    //
                    list.addVarBind(randomVar);
                    
                    // Compute the Priv key change Oid.
                    final SnmpOid privKeyChangeOid = 
                        new SnmpOid("1.3.6.1.6.3.15.1.2.2.1.9"
                                    +"."+userIndex);
                    
                    // Calculate Privacy delta
                    // This is where the SnmpUsmKeyHandler is to be used.
                    // Passing it the authentication protocol name it will
                    // compute the various keys.
                    //
                    // From the old password, compute a key.
                    //
                    final byte[] oldKey = 
                        handler.password_to_key(authProtocol, privPasswd);
                    
                    // println("OLD KEY :" + SnmpTools.binary2ascii(oldKey));

                    // Localize with the engine Id this old key.
                    // 16 is the size of a DES priv algorithm key.
                    // Only DES is supported in this example.
                    //
                    final byte[] oldKul = 
                        handler.localizePrivKey(authProtocol, oldKey, 
                                                engineId, 16);
                    
                    // println("OLD KUL :" + SnmpTools.binary2ascii(oldKul));

                    // From the new password, compute a key.
                    // 
                    byte[] newKey = handler.password_to_key(authProtocol, 
                                                            privNewPasswd);

                    // println("NEW KEY :" + SnmpTools.binary2ascii(newKey));

                    // Localize with the engine Id this new key.
                    // 16 is the size of a DES priv algorithm key.
                    // Only DES is supported in this example.
                    //
                    final byte[] newKul = 
                        handler.localizePrivKey(authProtocol, newKey, 
                                                engineId, 16);

                    // println("NEW KUL :" + SnmpTools.binary2ascii(newKul));

                    // Compute the privacy delta.
                    //
                    final byte[] del = 
                        handler.calculatePrivDelta(authProtocol, oldKul, 
                                                   newKul,
                                                   hexRandom.getBytes(),16);
                    
                    // println("DELTA :" + SnmpTools.binary2ascii(del));

                    // Put the delta in a VarBind.
                    //
                    final SnmpString delta = new SnmpString(del);
                    final SnmpVarBind privKeyChange = 
                        new SnmpVarBind(privKeyChangeOid, delta);

                    // Add the Privacy key change to the VarBindList
                    //
                    list.addVarBind(privKeyChange);

                    // Do the request.
                    //
                    doRequest(list, false);
                    println("***** Priv key change DONE.");
                }
                step = 2;
            }

            // Authentication key change (step 2)
            //
            if (step == 2) {
                // Auth Key changes is needed
                if (securityLevel.equals("authPriv") ||
                    securityLevel.equals("authNoPriv")) {
                    println("***** Doing Auth key change");
                    
                    // Construct the VarBindList.
                    // 
                    final SnmpVarBindList list = 
                        new SnmpVarBindList("Auth key change varbind list");

                    // Get the spinLock and add the result to VarBindList
                    //
                    final SnmpVarBindList result = doRequest(spinlist,true);
                    list.addVarBind(result.getVarBindAt(0));

                    
                    // Handle the random part of the Auth key change.
                    // Make use of SnmpString.HexToChar method.
                    // This is a useful method that converts an hexadecimal
                    // binary string in a String.
                    // Once converted, the String bytes contains the 
                    // binary value, not the ascii one.
                    //
                    final String hexRandom = 
                        SnmpString.HexToChar(authRandom);
                    final SnmpString random = 
                        new SnmpString(hexRandom.getBytes());
                    
                    // Compute the random Oid.
                    //
                    final SnmpOid randomOid = 
                        new SnmpOid("1.3.6.1.6.3.15.1.2.2.1.11"+"."+
                                    userIndex);

                    // Create the random varbind
                    //
                    final SnmpVarBind randomVar = 
                        new SnmpVarBind(randomOid, random);
                   
                    // Add the random in the VarBindList
                    //
                    list.addVarBind(randomVar);

                    // Compute the Authentication key change Oid.
                    //
                    final SnmpOid authKeyChangeOid = 
                        new SnmpOid("1.3.6.1.6.3.15.1.2.2.1.6"
                                    +"."+userIndex);
                    
                    
                    // Compute Auth delta.
                    // This is where the SnmpUsmKeyHandler is to be used.
                    // Passing it the authentication protocol name it will
                    // compute the various keys.
                    //

                    // From the old password, compute a key.
                    //
                    final byte[] oldKey = 
                        handler.password_to_key(authProtocol, authPasswd);

                    // Localize with the engine Id this old key.
                    //
                    final byte[] oldKul = 
                        handler.localizeAuthKey(authProtocol, oldKey, 
                                                engineId);
                    
                    // From the new password, compute a key.
                    //
                    final byte[] newKey = 
                        handler.password_to_key(authProtocol,authNewPasswd);

                    // Localize with the engine Id this new key.
                    //
                    final byte[] newKul = 
                        handler.localizeAuthKey(authProtocol, newKey, 
                                                engineId);

                    // Compute the auth delta.
                    //
                    final byte[] del = 
                        handler.calculateAuthDelta(authProtocol, oldKul, 
                                                   newKul, 
                                                   hexRandom.getBytes());

                    // Make the delta a VarBind.
                    //
                    final SnmpString delta = new SnmpString(del);
                    final SnmpVarBind authKeyChange = 
                        new SnmpVarBind(authKeyChangeOid, delta);
                    
                    // Add the Auth key change in the VarBindList
                    //
                    list.addVarBind(authKeyChange);
                    
                    // Do the request.
                    //
                    doRequest(list, false);
                    println("***** Auth key change DONE.");
                }
                step = 3;
            }

            // If no Privacy, set the privacy algorithm of the new user 
            // to usmNoPrivProtocol (step 3)
            //
            if (step == 3) {
                // If authNoPriv, set to no algo for priv algo.
                //
                if (securityLevel.equals("authNoPriv") || 
                    securityLevel.equals("noAuthNoPriv") ) {
                    println("***** Setting priv algo to noAlgo");

                    // Create the varbindlist
                    //
                    final SnmpVarBindList list = 
                        new SnmpVarBindList("No priv algo varbind list");

                    // Access spinLock and add it to the list.
                    //
                    final SnmpVarBindList result = doRequest(spinlist,true);
                    list.addVarBind(result.getVarBindAt(0));

                    // Compute privacy algorithm oid.
                    // 
                    final SnmpOid algoOid = 
                        new SnmpOid("1.3.6.1.6.3.15.1.2.2.1.8" + "."+
                                    userIndex);
                    
                    // Create privacy algorithm varbind.
                    // The value is the oid usmNoPrivProtocol.
                    //
                    final SnmpVarBind algo = 
                        new SnmpVarBind(algoOid,
                                        new SnmpOid("1.3.6.1.6.3.10.1.2.1"));

                    // Add the varbind to the list.
                    //
                    list.addVarBind(algo);

                    // Do the request.
                    //
                    doRequest(list, false);
                    println("***** Setting priv algo to noAlgo DONE.");
                }
                step = 4;
            }
            
            // If no Authentication, set the authentication algorithm 
            // of the new user to usmNoAuthProtocol (step 4)
            //
            if (step == 4) {
                // If noAuthNoPriv, set to no algo for auth algo.
                //
                if (securityLevel.equals("noAuthNoPriv")) {
                    println("***** Setting auth algo to noAlgo.");
                    // Create the varbindlist
                    //
                    final SnmpVarBindList list = 
                        new SnmpVarBindList("No auth algo varbind list");

                    // Access spinLock and add it to the list.
                    //
                    final SnmpVarBindList result = doRequest(spinlist,true);
                    list.addVarBind(result.getVarBindAt(0));
                    
                    // Compute Auth algorithm oid.
                    //
                    final SnmpOid algoOid = 
                        new SnmpOid("1.3.6.1.6.3.15.1.2.2.1.5" +"."+
                                    userIndex);
                    
                    // Create auth algo varbind.
                    // The value is the oid usmNoAuthProtocol.
                    //
                    SnmpVarBind algo = 
                        new SnmpVarBind(algoOid,
                                        new SnmpOid("1.3.6.1.6.3.10.1.1.1"));

                    // Add varbind to the list.
                    //
                    list.addVarBind(algo);

                    // Do the request.
                    //
                    doRequest(list, false);
                    println("***** Setting auth algo to noAlgo DONE.");
                }
                step = 5;
            }

            // The last step, activate the row (step 5).
            //
            if (step == 5) {

                // Set the row status to active
                // 
                println("***** Setting row status to active.");

                // Create the varbindlist
                //
                final SnmpVarBindList list = 
                    new SnmpVarBindList("Active row status varbind list");
                
                // Access spinLock and add it to the list.
                //
                final SnmpVarBindList result = doRequest(spinlist, true);
                list.addVarBind(result.getVarBindAt(0));
                
                // Create the active value.
                //
                final SnmpInt status = new SnmpInt(EnumRowStatus.active);
                
                // Create the row status varbind and add it to the list
                //
                final SnmpVarBind rowStatus = 
                    new SnmpVarBind(statusOid, status);
                list.addVarBind(rowStatus);

                // Do the rerquest
                //
                doRequest(list, false);

                //
                // OVER. EVERYTHING WAS OK.....
                //
                //
                println("***** Setting row status to active DONE.");
                println("\n***** SUCCESSFULLY CREATED NEW ROW IN AGENT" +
                        " FOR USER : [" + userName + "]*****");
                
                // Sanity check. Make a get request with the new user.
                //
                print("\nSend sanity check? Your manager.security file MUST " +
                      "contain the currently created user (press return to " +
                      "do it, \"no\" to skip):");
                String doit = in.readLine();
                if (!doit.equals("no")) {
                    println("\n\n***** Making a sanity check get request" +
                            " on the spin lock with the newly created user ["+ 
                            userName + "].");               
                    boolean discoveryFlag = false;
                    step = 6;
                    final SnmpUsmParameters sec = 
                        new SnmpUsmParameters(engine, userName);
                    
                    if (securityLevel.equals("authPriv")) {
                        discoveryFlag = true;
                        sec.setSecurityLevel(SnmpDefinitions.authPriv);
                    }
                    if (securityLevel.equals("authNoPriv")) {
                        discoveryFlag = true;
                        sec.setSecurityLevel(SnmpDefinitions.authNoPriv);
                    }
                    if (securityLevel.equals("noAuthNoPriv"))
                        sec.setSecurityLevel(SnmpDefinitions.noAuthNoPriv);
                        
                    sec.setContextEngineId(agent.getEngineId().getBytes());
                    
                    // Cache the current ones.
                    //
                    final SnmpParams old = agent.getParams();
                    
                    try {
                        // Set the new params.
                        //
                        agent.setParams(sec);
                        
                        // Ensure that the request will be in time window.
                        //
                        if (discoveryFlag)
                            agent.processUsmTimelinessDiscovery();
                        
                        // Do the spinlock get request
                        //
                        final SnmpVarBindList sanityCheckResult = 
                            doRequest(spinlist, true);
                        final SnmpVarBind spinVarBind = 
                            sanityCheckResult.getVarBindAt(0);
                        final int newSpinLockValue = 
                            ((SnmpInt)(spinVarBind.getSnmpValue())).intValue();
                        println("\n SANITY CHECK SUCCESSFUL, " + 
                                "SPIN LOCK VALUE IS: "+
                                newSpinLockValue);
                    } catch (Exception e) {
                        println("\n SANITY CHECK FAILED : " + e.toString());
                        println("To allow the request to be sent, the newly " 
                                + "created user [" + userName + "]");
                        println("must also be in the manager.security file.");
                        println("Add the following to manger security file:");
                        
                        String common = "userEntry=" + engineId +"," +
                            userName + "," + 
                            userName;
                        StringBuffer finalLine = new StringBuffer(common);
                        
                        if (securityLevel.equals("authPriv"))
                            finalLine.append("," + 
                                             authProtocol + "," + 
                                             authNewPasswd + "," + 
                                             "usmDESPrivProtocol," +
                                             privNewPasswd);
                        else    
                            if (securityLevel.equals("authNoPriv"))
                                finalLine.append("," + 
                                                 authProtocol + "," + 
                                                 authNewPasswd);
                        
                        println(finalLine.toString());
                    }
                                // Set back the current parameters
                    //
                    agent.setParams(old);
                }
             
                step = 0;
                undo = true;
            }

        } catch (Exception e) {
            println("\n***** FAILED TO CREATE NEW ROW IN AGENT FOR USER:" +
                    " [" + userName + "] Exception raised by agent [" + e +"]");
            e.printStackTrace();
            if (!undo) {
                println("\t * User created. In agent. Failed step is: " + 
                        step2String(step));
                println("\tDo you want to restart from the failed step? " +
                        " (yes, no) :");
                try {
                    final String ans = in.readLine();
                    if (ans.equals("no")) {
                        step = 0;
                        undo = true;
                    }
                } catch (Exception x) {
                    println("Input failure");
                    java.lang.System.exit(0);
                }
            }
        }
    }
    
    private String step2String(int step) {
        switch(step) {
        case 0: return "Row creation set(spin,status,cloneFrom)";
        case 1: return "Privacy key change";
        case 2: return "Authentication key change";
        case 3: return "Set privacy algorithm to none";
        case 4: return "Set authentication algorithm to none";
        case 5: return "Row activation";
        case 6: return "Sanity check, get spin lock failed";
        default:
        }
        return null;
    }
    
    public static void  usage() {
        println("java -Djdmk.security.file=manager.security" +
                " CreateUsmMibUser <name> <securityLevel> <host> <port>");
        println("where");
        println("\t-name: User name to use when sending requests to agent.");
        println("\t-securityLevel: Security level to use when sending" +
                " requests to agent (\"authPriv\", \"authNoPriv\", " +
                "\"noAuthNoPriv\")");
        println("\t-host: hostname or IP address of the SNMP V3 agent" +
                " you want to query.");
        println("\t-port: port number to use.");
    }

    /**
     * print/println stuff...
     */

    private final static void println(String msg) {
        java.lang.System.out.println(msg);
    }

    private final static void print(String msg) {
        java.lang.System.out.print(msg);
    }
}
